SystemOrganization addCategory: #'Paillier-Model'!
SystemOrganization addCategory: #'Paillier-Tests'!

Object subclass: #PaillierKeyPairGenerator
	instanceVariableNames: 'bits p q g lambda n x phinsquared nsquared'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Paillier-Model'!

!PaillierKeyPairGenerator commentStamp: 'NicolasVanderAuwera 3/10/2012 15:52' prior: 0!
Key pair generator for the Paillier Cryptosystem. I am based on the algorithm proposed by Mads Johan Jurik. See http://acsc.cs.utexas.edu/libpaillier/.!

!PaillierKeyPairGenerator class methodsFor: 'instance creation' stamp: 'NicolasVanderAuwera 2/27/2012 07:38'!
bits: anInteger
	^ self new bits: anInteger! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/12/2012 07:53'!
bits
	^ bits! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 2/27/2012 07:39'!
bits: anObject
	bits := anObject! !

!PaillierKeyPairGenerator methodsFor: 'private' stamp: 'NicolasVanderAuwera 2/27/2012 07:40'!
computePrimes
	p := self findPrime.
	q := self findPrime! !

!PaillierKeyPairGenerator methodsFor: 'private' stamp: 'NicolasVanderAuwera 3/10/2012 14:35'!
computeX
	| eeu |
	x := self g raisedTo: lambda modulo: n * n.
	x := x - 1.
	x := x / n.
	eeu := ExtendedEuclidAlgorithm with: x with: n.
	x := eeu inverseModulo! !

!PaillierKeyPairGenerator methodsFor: 'private' stamp: 'NicolasVanderAuwera 3/12/2012 07:51'!
findPrime
	^ PrimesFinder primeOfSize: bits // 2! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/10/2012 11:15'!
g
	g ifNil: [ g := self n + 1 ].
	^ g! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 5/6/2012 15:08'!
lambda
	^ lambda ifNil: [ lambda := (self p - 1) * (self q - 1) / 2 ]! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/10/2012 11:15'!
n
	n ifNil: [ n := self p * self q ].
	^ n! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/28/2012 15:41'!
nsquared
	nsquared ifNil: [ nsquared := self n * self n ].
	^ nsquared! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 5/9/2012 10:22'!
p
	p ifNil: [ self computePrimes ].
	^ p! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/28/2012 15:29'!
phinsquared
	phinsquared ifNil: [ phinsquared := n * ((self p - 1) * (self q - 1)) ].
	^ phinsquared! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/11/2012 09:18'!
privateKey
	^ PaillierPrivateKey lambda: self lambda x: self x p: self p q: self q! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/10/2012 11:16'!
publicKey
	^ PaillierPublicKey n: self n! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 5/9/2012 10:22'!
q
	q ifNil: [ self computePrimes ].
	^ q! !

!PaillierKeyPairGenerator methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/10/2012 14:25'!
x
	x ifNil: [ self computeX ].
	^ x! !

Object subclass: #PaillierPrivateKey
	instanceVariableNames: 'lambda x p q blindingKey n phinsquared rand'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Paillier-Model'!

!PaillierPrivateKey commentStamp: 'NicolasVanderAuwera 3/10/2012 15:52' prior: 0!
Private key of the Paillier's Cryptosystem.!

!PaillierPrivateKey class methodsFor: 'instance creation' stamp: 'NicolasVanderAuwera 3/11/2012 09:18'!
lambda: int1 x: int2 p: int3 q: int4
	^ self new
		lambda: int1;
		x: int2;
		p: int3;
		q: int4;
		yourself! !

!PaillierPrivateKey methodsFor: 'process' stamp: 'NicolasVanderAuwera 3/10/2012 14:31'!
crypt: c with: pubKey
	| m |
	m := c raisedTo: lambda modulo: pubKey nsquared.
	m := m - 1.
	m := m / pubKey n.
	m := m * x.
	^ m \\ pubKey n! !

!PaillierPrivateKey methodsFor: 'process' stamp: 'NicolasVanderAuwera 3/29/2012 07:16'!
decrypt: aCollection with: pk
	"Decrypt each cell of the collection separately."

	^ aCollection collect: [ :cipher | self crypt: cipher with: pk ]! !

!PaillierPrivateKey methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/11/2012 10:32'!
lambda
	^ lambda! !

!PaillierPrivateKey methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 2/27/2012 08:00'!
lambda: anObject
	lambda := anObject! !

!PaillierPrivateKey methodsFor: 'private' stamp: 'NicolasVanderAuwera 3/11/2012 10:40'!
n
	n ifNil: [ n := p * q ].
	^ n! !

!PaillierPrivateKey methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/11/2012 09:18'!
p: anInteger
	p := anInteger! !

!PaillierPrivateKey methodsFor: 'private' stamp: 'NicolasVanderAuwera 3/11/2012 10:41'!
phinsquared
	phinsquared ifNil: [ phinsquared := self n * ((p - 1) * (q - 1)) ].
	^ phinsquared! !

!PaillierPrivateKey methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/11/2012 09:18'!
q: anInteger
	q := anInteger! !

!PaillierPrivateKey methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/11/2012 10:32'!
x
	^ x! !

!PaillierPrivateKey methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/10/2012 14:35'!
x: anInteger
	x := anInteger! !

Object subclass: #PaillierPublicKey
	instanceVariableNames: 'n g'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Paillier-Model'!

!PaillierPublicKey commentStamp: 'NicolasVanderAuwera 3/10/2012 15:52' prior: 0!
Public key of the Paillier's Cryptosystem.!

!PaillierPublicKey class methodsFor: 'instance creation' stamp: 'NicolasVanderAuwera 3/10/2012 11:16'!
n: anInteger
	^ self new
		n: anInteger;
		yourself! !

!PaillierPublicKey methodsFor: 'process' stamp: 'NicolasVanderAuwera 3/29/2012 07:31'!
crypt: m
	| r c x |
	self assert: m < n.
	r := SecureRandom new nextInt: n.
	c := self g raisedTo: m modulo: self nsquared.
	x := r raisedTo: self n modulo: self nsquared.
	^ c * x \\ self nsquared! !

!PaillierPublicKey methodsFor: 'process' stamp: 'NicolasVanderAuwera 3/29/2012 07:17'!
encrypt: aCollection
	"Encrypt each cell of the collection separately."

	^ aCollection collect: [ :msg | self crypt: msg ]! !

!PaillierPublicKey methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/10/2012 11:17'!
g
	g ifNil: [ g := self n + 1 ].
	^ g! !

!PaillierPublicKey methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 2/27/2012 11:25'!
n
	^ n! !

!PaillierPublicKey methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 2/27/2012 08:02'!
n: anObject
	n := anObject! !

!PaillierPublicKey methodsFor: 'accessing' stamp: 'NicolasVanderAuwera 3/10/2012 11:25'!
nsquared
	^ self n raisedTo: 2! !

TestCase subclass: #PaillierKeyPairGeneratorTest
	instanceVariableNames: 'pk sk gen'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Paillier-Tests'!

!PaillierKeyPairGeneratorTest methodsFor: 'cryptography' stamp: 'NicolasVanderAuwera 2/27/2012 12:50'!
elementOfMultiplicativeGroup: n
	| g r |
	r := SecureRandom new.
	[ 
	g := r nextRandom160.
	(g gcd: n) = 1 ] whileFalse.
	^ g! !

!PaillierKeyPairGeneratorTest methodsFor: 'running' stamp: 'NicolasVanderAuwera 3/28/2012 15:36'!
setUp
	| generator |
	generator := PaillierKeyPairGenerator bits: 1024.
	pk := generator publicKey.
	sk := generator privateKey! !

!PaillierKeyPairGeneratorTest methodsFor: 'tests' stamp: 'NicolasVanderAuwera 3/10/2012 14:40'!
testEncDec
	| m c |
	m := 56789.
	c := pk crypt: m.
	self assert: (sk crypt: c with: pk) = m! !

!PaillierKeyPairGeneratorTest methodsFor: 'tests' stamp: 'NicolasVanderAuwera 3/29/2012 07:20'!
testEncryptDecrypt
	| array c |
	array := #(56789 5263 8923 9192).
	c := pk encrypt: array.
	self assert: (sk decrypt: c with: pk) = array! !

!PaillierKeyPairGeneratorTest methodsFor: 'tests' stamp: 'NicolasVanderAuwera 5/9/2012 10:13'!
testHomomorphism
	| vote1 vote2 c1 c2 results |
	vote1 := #(1 0 0 0).
	vote2 := #(0 0 0 1).
	c1 := pk encrypt: vote1.
	c2 := pk encrypt: vote2.
	results := c1 * c2.
	self assert: (sk decrypt: results with: pk) = (vote1 + vote2)
	! !
